#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <ctype.h>
#include <errno.h>

#include <assert.h>

#include "sh.h"
#include "shio.h"
#include "shvar.h"
#include "util.h"
#include "parser.h"
#include "charset.h"

jmp_buf		parse_error;

/*
 * The command language implemented by this shell is very simple.
 *
 * The processing phases of a simple command:
 *
 * - NO backquote expansion or eval (if we handled backquoted text, it
 * would be processed separately w/ the same stages below.
 *
 * - NO history expansion
 *
 * - quote separation
 *
 * - dollar expansion of shell/environment variables.
 *
 * - separation into words.  (variables can expand to multiple words)
 *
 * - redirection (this allows $n>($p) to work), but the '>' must have
 * been there originally (not part of dollar expansion output).
 *
 * - execution.
 *
 * Test cases
 *
 * given lvar="a b c", svar="s", the expansions, with "/" separating
 * the words:
 *
 *  $lvar -> a/b/c
 *  x "$lvar" -> x/a b c
 *  $svar"$lvar" -> sa b c
 *  $lvar$svar -> a/b/cs
 *  $lvar$lvar -> a/b/ca/b/c
 *  $lvar'$lvar' ->a/b/c$lvar
 */

void		lexer_constr(struct lexer	*lp,
			     FILE		*str)
{
	lp->stream = str;
	lp->print_prompt = 1;	/* should be only if interactive shell */
	lp->state = LEXER_NORM;
	lp->word_dest = rnone;
	lp->direction = rin;
	lp->target = 0;
	strbuf_constr(&lp->vname);
}

struct lexer	*lexer_constr_ptr(FILE	*str)
{
	struct lexer	*lp;

	lp = safe_malloc(sizeof *lp);
	lexer_constr(lp,str);
	return lp;
}

void		lexer_destr(struct lexer	*lp)
{
	strbuf_destr(&lp->vname);
}

void		lexer_destr_ptr(struct lexer	*lp)
{
	lexer_destr(lp);
	free(lp);
}

void 		cmd_constr(struct cmd	*cmdp)
{
	cmdp->rop = 0;
	cmdp->background = 0;
	dynbuf_constr(&cmdp->words);
	dynbuf_constr(&cmdp->vars);
}

struct cmd	*cmd_constr_ptr(void)
{
	struct cmd	*cmdp;

	cmdp = safe_malloc(sizeof *cmdp);
	cmd_constr(cmdp);

	return cmdp;
}

struct redir_op	*redir_op_constr_ptr(void)
{
	struct redir_op	*rop = safe_malloc(sizeof *rop);
	rop->next = 0;
	rop->src_type = rnone;
	return rop;
}

void		redir_op_destr_ptr(struct redir_op	*rop)
{
	if (rop) {
		redir_op_destr_ptr(rop->next);
		switch (rop->src_type) {
		case rfile:
		case rpipe:
		case rboth:
			free(rop->src.name);
		case rnone:
		case rdesc:
		case rclose:
			;
		}
		free(rop);
	}
}

void		cmd_destr(struct cmd	*cmdp)
{
	redir_op_destr_ptr(cmdp->rop);
	dynbuf_destr(&cmdp->words);
	dynbuf_destr(&cmdp->vars);
}

void		cmd_destr_ptr(struct cmd	*cmdp)
{
	cmd_destr(cmdp);
	free(cmdp);
}

char		*is_varset(char		*word)
{
	char	*p, ch;

	p = word;
	if (!(lexclass[0xff & *p++] & VNANY)) return 0;
	while ((ch = *p++) != 0 && ch != '=') {
		if (!(lexclass[0xff & ch] & (VNANY|VNREST))) return 0;
	}
	if (ch != '=') return 0;

	return p;
}

void		cmd_add_word(struct cmd	*cmdp,
			     char	*w)
{
	dynbuf_add_elt(&cmdp->words,w);
}

void		cmd_add_var(struct cmd	*cmdp,
			    char	*v)
{
	dynbuf_add_elt(&cmdp->vars,v);
}

void		cmd_add_rop(struct cmd		*cmdp,
			    struct redir_op	*op)
{
	op->next = cmdp->rop;
	cmdp->rop = op;
}

static void	xfer_word(struct lexer		*lp,
			  struct cmd		*cmdp,
			  struct strbuf		*sb)
{
	char		*word = 0, *p, ch;
	struct redir_op	*op = 0;

	if (strbuf_nchars(sb) || lp->word_dest == rclose) {
		/*
		 * if no characters and it's not rclose, this is just
		 * input field separators that occur repeatedly or at
		 * the beginning/end of the input line.
		 */
		if (lp->word_dest != rclose) {
			strbuf_add_ch(sb,'\0');
			word = strbuf_take_buf(sb);
		}
		if (lp->word_dest != rnone) {
			op = redir_op_constr_ptr();
			op->dir = lp->direction;
			op->src_type = lp->word_dest;
			op->target_desc = lp->target;
		}
		switch (lp->word_dest) {
		case rboth:
			desc_printf(2,SHNAME ": rboth in xfer_word\n");
			abort();
		case rnone:
			if (!dynbuf_nelt(&cmdp->words) && is_varset(word) != 0) {
				cmd_add_var(cmdp,word);
			} else {
				cmd_add_word(cmdp,word);
			}
			break;
		case rdesc:
			/*
			 * if all digits, use as a descriptor;
			 * otherwise it is a file name for both
			 * standard input and standard error.
			 */
			for (p = word; (ch = *p++) != 0;)
				if (!isdigit(0xff & ch)) {
					op->src_type = rboth;
					goto rfile_processing;
				}
			op->src.desc = atoi(word);
			free(word);
			cmd_add_rop(cmdp,op);
			break;
		case rfile:
		rfile_processing:
			op->src_type = rfile;
			/* fall through */
		case rpipe:
			op->src.name = word;
			/* fall through */
		case rclose:
			cmd_add_rop(cmdp,op);
			break;
		}
		lp->word_dest = rnone;
		lp->target = 0;
	}
}

static void	abort_line(struct lexer		*in_lexer)
{
	int	ch;

	while ((ch = getc(in_lexer->stream)) != EOF && ch != '\n')
		;
}

void		xfer_value_norm(struct lexer	*in_lexer,
				struct cmd	*cmdp,
				struct strbuf	*sbp)
{
	char		*vname;
	const char	*vvalue;
	int		ch;

	/* transfer the value of the variable */
	vname = strbuf_take_buf(&in_lexer->vname);
	vvalue = dollar_value(vname);
	if (!vvalue) {
		desc_printf(2,SHNAME ": variable %s not set\n",vname);
		longjmp(parse_error,1);
	}
	while ((ch = *vvalue++) != 0) {
		if (lexclass[ch] & IFS) {
			xfer_word(in_lexer,cmdp,sbp);
		} else {
			strbuf_add_ch(sbp,ch);
		}
	}
	free(vname);
}

void		xfer_value_quoted(struct lexer	*in_lexer,
				  struct cmd	*cmdp,
				  struct strbuf	*sb)
{
	char		*vname;
	const char	*vvalue;
	int		ch;

	/* transfer the value of the variable */
	vname = strbuf_take_buf(&in_lexer->vname);
	vvalue = dollar_value(vname);
	if (!vvalue) {
		desc_printf(2,SHNAME ": variable %s not set\n",vname);
		longjmp(parse_error,1);
	}
	while ((ch = *vvalue++) != 0) {
		strbuf_add_ch(sb,ch);
	}
	free(vname);
}

/*
 * get command until semicolon or newline
 */
struct cmd	*get_simple_cmd(struct lexer	*in_lexer)
{
	struct cmd	*cmdp;
	struct strbuf	sb;
	int		ch;
	int		target_specified;

	cmdp = 0;
	strbuf_constr(&sb);
	if (setjmp(parse_error)) {
		/* clean up */
		abort_line(in_lexer);
		if (cmdp) cmd_destr_ptr(cmdp);
		strbuf_destr(&sb);

		in_lexer->print_prompt = 1;
		/* fall through: continue in case of a parse error */
	}
	cmdp = cmd_constr_ptr();
	strbuf_constr(&sb);/* XXX: double constr */

	in_lexer->state = LEXER_NORM;
	in_lexer->target = 0;
	target_specified = 0;
	do {
		if (in_lexer->print_prompt) {
			printf("Y$ "); fflush(stdout);
			in_lexer->print_prompt = 0;
		}
		do {
			errno = 0;
			ch = getc(in_lexer->stream);
		} while (ferror(in_lexer->stream) && errno == EINTR && (clearerr(in_lexer->stream), 1));

		switch (in_lexer->state) {
		case LEXER_NORM_LESS_AMP:
		case LEXER_NORM_GTR_AMP:
			in_lexer->state = LEXER_NORM;
			if (ch == '-') {
				in_lexer->word_dest = rclose;
				xfer_word(in_lexer,cmdp,&sb);
				break;
			}
			/* fall through */
		case LEXER_NORM:
		norm_processing:
			switch (ch) {
			case '&':
				cmdp->background = 1;
				goto end_cmd;
			case '\n':
				in_lexer->print_prompt = 1;
				/* fall through */
			case ';':
			end_cmd:
				xfer_word(in_lexer,cmdp,&sb);
				goto done;
			case '\'':
				in_lexer->state = LEXER_SGL; break;
			case '"':
				in_lexer->state = LEXER_DBL; break;
			case '\\':
				in_lexer->state = LEXER_NORM_BSLASH; break;
			case '$':
				in_lexer->state = LEXER_NORM_DLR;
				strbuf_clear(&in_lexer->vname);
				/* noop after destr */
				break;
			default:
				/* lexclass handles EOF */
				if (lexclass[ch] & REDIR) {
					if (!target_specified) {
						/*
						 * echo foo5>bar
						 * puts foo5 into bar
						 */
						xfer_word(in_lexer,cmdp,&sb);
						in_lexer->target = (ch == '>');
					}
					target_specified = 0;
					strbuf_clear(&sb);
					in_lexer->state = (ch == '>') ? LEXER_NORM_GTR : LEXER_NORM_LESS;
				} else if (lexclass[ch] & IFS) {
					xfer_word(in_lexer,cmdp,&sb);
				} else if (in_lexer->target >= 0 && isdigit(ch)) {
					if (in_lexer->word_dest == rnone) {
						in_lexer->target *= 10;
						in_lexer->target += ch - '0';
						target_specified = 1;
					}
					/*
					 * this may be echo 1 rather than
					 * echo 1> foo
					 */
					strbuf_add_ch(&sb,ch);
				} else {
					if (in_lexer->word_dest == rnone)
						in_lexer->target = -1;
					strbuf_add_ch(&sb,ch);
				}
			}
			break;
		case LEXER_NORM_BSLASH:
			in_lexer->target = -1;
			if (ch != EOF)
				strbuf_add_ch(&sb,ch);
			in_lexer->state = LEXER_NORM;
			break;
		case LEXER_NORM_DLR:
			in_lexer->target = -1;
			if (lexclass[ch] & VNANY) {
				strbuf_add_ch(&in_lexer->vname,ch);
				in_lexer->state = LEXER_NORM_DLR2;
			} else if (lexclass[ch] & VNSPEC) {
				strbuf_add_ch(&in_lexer->vname,ch);
				strbuf_add_ch(&in_lexer->vname,'\0');
				in_lexer->state = LEXER_NORM;
				xfer_value_norm(in_lexer,cmdp,&sb);
			} else {
				desc_printf(2,SHNAME ": vars must start with ALPHA\n");
				longjmp(parse_error,1);
			}
			break;
		case LEXER_NORM_DLR2:
			in_lexer->target = -1;
			/* look for end-of-word and xfer shvar value */
			if (lexclass[ch] & (VNREST|VNANY)) {
				strbuf_add_ch(&in_lexer->vname,ch);
			} else {
				in_lexer->state = LEXER_NORM;
				ungetc(ch,in_lexer->stream);
				strbuf_add_ch(&in_lexer->vname,'\0');
				xfer_value_norm(in_lexer,cmdp,&sb);
			}
			break;
		case LEXER_DBL:
			in_lexer->target = -1;
			/* look for matching double quotes */
			switch (ch) {
			case '\\':
				in_lexer->state = LEXER_DBL_BSLASH;
				break;
			case '"':
				in_lexer->state = LEXER_NORM;
				break;
			case '$':
				in_lexer->state = LEXER_DBL_DLR;
				break;
			default:
				if (ch != EOF)
					strbuf_add_ch(&sb,ch);
				else {
					desc_printf(2,SHNAME ": EOF during double-quotes processing\n");
					return 0;
				}
				break;
			}
			break;
		case LEXER_DBL_BSLASH:
			in_lexer->target = -1;
			if (ch != EOF)
				strbuf_add_ch(&sb,ch);
			in_lexer->state = LEXER_DBL;
			break;
		case LEXER_DBL_DLR:
			in_lexer->target = -1;
			if (lexclass[ch] & VNANY) {
				strbuf_add_ch(&in_lexer->vname,ch);
				in_lexer->state = LEXER_DBL_DLR2;
			} else if (lexclass[ch] & VNSPEC) {
				strbuf_add_ch(&in_lexer->vname,ch);
				strbuf_add_ch(&in_lexer->vname,'\0');
				in_lexer->state = LEXER_DBL;
				xfer_value_quoted(in_lexer,cmdp,&sb);
			} else {
				desc_printf(2,SHNAME ": vars must start with ALPHA\n");
				longjmp(parse_error,1);
			}
			break;
		case LEXER_DBL_DLR2:
			in_lexer->target = -1;
			/* look for end-of-word and xfer shvar value */
			if (lexclass[ch] & (VNREST|VNANY)) {
				strbuf_add_ch(&in_lexer->vname,ch);
			} else {
				in_lexer->state = LEXER_DBL;
				ungetc(ch,in_lexer->stream);
				strbuf_add_ch(&in_lexer->vname,'\0');

				xfer_value_quoted(in_lexer,cmdp,&sb);
			}
			break;
		case LEXER_SGL:
			in_lexer->target = -1;
			/* look for matching single quote */
			if (ch == '\'') {
				in_lexer->state = LEXER_NORM;
			} else {
				if (ch != EOF)
					strbuf_add_ch(&sb,ch);
				else {
					desc_printf(2,SHNAME ": EOF during single quote processing\n");
					return 0;
				}
			}
			break;
		case LEXER_NORM_LESS:
			in_lexer->direction = rin;
			if (lexclass[ch] & IFS) {
				in_lexer->state = LEXER_NORM;
				in_lexer->word_dest = rfile;
				break;
			}
			switch (ch) {
			case EOF:
				desc_printf(2,SHNAME ": EOF in input-redirection specifier\n");
				longjmp(parse_error,1);
			case '&':
				in_lexer->word_dest = rdesc;
				in_lexer->state = LEXER_NORM_LESS_AMP;
				break;
			case '|':
				in_lexer->word_dest = rpipe;
				in_lexer->state = LEXER_NORM;
				break;
			default:
				/* filename sans whitespace. */
				in_lexer->word_dest = rfile;
				in_lexer->state = LEXER_NORM;
				goto norm_processing;
			}
			break;
		case LEXER_NORM_GTR:
			in_lexer->direction = rout;
			if (lexclass[ch] & IFS) {
				in_lexer->state = LEXER_NORM;
				in_lexer->word_dest = rfile;
				break;
			}
			switch (ch) {
			case EOF:
				desc_printf(2,SHNAME ": EOF in input-redirection specifier\n");
				longjmp(parse_error,1);
			case '&':
				in_lexer->word_dest = rdesc;
				in_lexer->state = LEXER_NORM_GTR_AMP;
				break;
			case '|':
				in_lexer->word_dest = rpipe;
				in_lexer->state = LEXER_NORM;
				break;
			case '>':
				in_lexer->direction = rfappend;
				in_lexer->word_dest = rfile;
				in_lexer->state = LEXER_NORM;
				break;
			default:
				/* filename sans whitespace */
				in_lexer->word_dest = rfile;
				in_lexer->state = LEXER_NORM;
				goto norm_processing;
			}
			break;
		}
	} while (ch != EOF);
	/* EOF null command */
	if (!dynbuf_nelt(&cmdp->words) && !dynbuf_nelt(&cmdp->vars)) {
		cmd_destr_ptr(cmdp);
		cmdp = 0;
	}
 done:
	strbuf_destr(&sb);
	return cmdp;
}
