#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <ctype.h>

extern char	**environ;

#include "sh.h"
#include "shio.h"
#include "util.h"
#include "parser.h"
#include "run.h"
#include "shvar.h"
#include "charset.h"

const int	debug = 0;

void	show_rops(struct redir_op	*rop)
{
	if (rop) {
		show_rops(rop->next);
		printf("%d",rop->target_desc);
		switch (rop->dir) {
		case rfappend:
			putchar('>');
			/* fall through */
		case rout:
			putchar('>');
			break;
		case rin:
			putchar('<');
			break;
		}
		switch (rop->src_type) {
		case rnone:
			printf("SHOULD NOT HAPPEN\n");
			break;
		case rfile:
			printf(" %s\n",rop->src.name);
			break;
		case rdesc:
			printf("&%d\n",rop->src.desc);
			break;
		case rpipe:
			printf("|%s\n",rop->src.name);
			break;
		case rclose:
			printf("&-\n");
			break;
		case rboth:
			printf(">& %s\n",rop->src.name);
			break;
		}
	}
}

int	do_cd(struct cmd	*cmdp)
{
	char	*homedir;

	if (dynbuf_nelt(&cmdp->words) > 2) {
		desc_printf(2,SHNAME ": chdir cannot have more than 1 argument\n");
		return 1;
	}
	if (dynbuf_nelt(&cmdp->words) == 1) {
		homedir=getenv("HOME");
		if (homedir) {
			homedir=strchr(homedir,'=');
			if (!homedir) {
				desc_printf(2,SHNAME ": HOME envariable malformed!\n");
				return 1;
			}
		} else {
			char		*loginname;
			struct passwd	*pw;

			if ((loginname = getlogin()) != 0) {
				pw = getpwnam(loginname);
			} else {
				pw = getpwuid(getuid());
			}
			homedir = pw->pw_dir;
		}
		if (chdir(homedir) == -1) {
			perror(SHNAME ": chdir");
			return 1;
		}
	} else {
		if (chdir((char *) dynbuf_get_elt(&cmdp->words,1)) == -1) {
			perror(SHNAME ": chdir");
			desc_printf(2,SHNAME ": cannot chdir to \"%s\"\n",(char *) dynbuf_get_elt(&cmdp->words,1));
			return 1;
		}
	}
	return 0;
}

int	do_exit(struct cmd	*cmdp)
{
	if (dynbuf_nelt(&cmdp->words) > 2) {
		desc_printf(2,SHNAME ": exit cannot have more than 1 argument\n");
		return 1;
	}
	if (dynbuf_nelt(&cmdp->words) == 1) {
		exit(atoi(dollar_value("?")));
	} else {
		exit(atoi((char *) dynbuf_get_elt(&cmdp->words,1)));
	}
	return 0;
}

int	ndesc = 0;
struct plist {
	struct plist	*next;
	char		*pname;
	int		d[2];
} *shpipes = 0;
int	**desc_used = 0;

void	pipe_init(void)
{
	ndesc = getdtablesize();
	if (ndesc == -1) {
		desc_printf(2,SHNAME ": panic - getdtablesize failed!?!\n");
		exit(1);
	}
	desc_used = calloc(ndesc,sizeof *desc_used);
}

void	desc_check(int	d)
{
	int	nd;

	if (!desc_used[d]) return;
	/*
	 * d is in use by a pipe.  sigh.
	 */
	nd = dup(*desc_used[d]);
	if (nd == -1) {
		desc_printf(2,SHNAME ": desc_check dup failed\n");
		exit(1);
	}
	*desc_used[d] = nd;
	desc_used[nd] = desc_used[d];
	desc_used[d] = 0;
	(void) close(d);
}

void	pipe_cleanup(void)
{
	struct plist	*p;

	for (p = shpipes; p; p = p->next) {
		if (p->d[0] != -1) (void) close(p->d[0]);
		if (p->d[1] != -1) (void) close(p->d[1]);
	}
}

int	do_pipe(struct cmd	*cmdp)
{
	struct plist	**ppp, *pp;

	if (!ndesc) pipe_init();
	if (dynbuf_nelt(&cmdp->words) != 2) {
		desc_printf(2,SHNAME ": pipe must have exactly one pipename argument\n");
		return 1;
	}
	for (ppp = &shpipes; (pp = *ppp) != 0; ppp = &pp->next) {
		if (!strcmp(pp->pname,dynbuf_get_elt(&cmdp->words,1))) {
			if (pp->d[0] != -1) {
				(void) close(pp->d[0]);
				desc_used[pp->d[0]] = 0;
			}
			if (pp->d[1] != -1) {
				(void) close(pp->d[1]);
				desc_used[pp->d[1]] = 0;
			}
			if (pipe(pp->d) == -1) {
				perror(SHNAME ": pipe");
				desc_printf(2,SHNAME ": could not create shell pipe \"%s\"\n",dynbuf_get_elt(&cmdp->words,1));
				*ppp = pp->next;
				free(pp);
				return 1;
			}
			desc_used[pp->d[0]] = &pp->d[0];
			desc_used[pp->d[1]] = &pp->d[1];
			return 0;
		}
	}
	pp = safe_malloc(sizeof *pp);
	if (pipe(pp->d) == -1) {
		perror(SHNAME ": pipe");
		desc_printf(2,SHNAME ": could not create shell pipe \"%s\"\n",dynbuf_get_elt(&cmdp->words,1));
		free(pp);
		return 1;
	}
	desc_used[pp->d[0]] = &pp->d[0];
	desc_used[pp->d[1]] = &pp->d[1];
	pp->pname = dupstr(dynbuf_get_elt(&cmdp->words,1));
	pp->next = shpipes;
	shpipes = pp;

	return 0;
}

int	do_pclose(struct cmd	*cmdp)
{
	int		nelt;
	struct plist	**ppp, *pp;
	char		*arg;
	int		ch, which = 7; /* bits: silent, write, read */

	if ((nelt = dynbuf_nelt(&cmdp->words)) != 2 && nelt != 3) {
		desc_printf(2,SHNAME ": pipe command has pipename and optional descriptor selector argument\n");
		return 1;
	}
	if (nelt == 3) {
		arg = dynbuf_get_elt(&cmdp->words,2);
		if (isdigit(0xff & arg[0])) {
			which = atoi(dynbuf_get_elt(&cmdp->words,2));
		} else {
			which = 0;
			while ((ch = *arg++) != 0) switch (ch) {
			case 'r':       which |= 1; break;
			case 'w':       which |= 2; break;
			default:
				desc_printf(2,SHNAME ": pipe optional descriptor selector argument must be 'r', 'w', or a number from 0 to 3.\n");
				return 1;
			}
		}
	}
	for (ppp = &shpipes; (pp = *ppp) != 0; ppp = &pp->next) {
		if (!strcmp(pp->pname,dynbuf_get_elt(&cmdp->words,1))) {
			if (which & 1) {
				if (pp->d[0] == -1) {
					if ((which & 4) == 0) {
						desc_printf(2,SHNAME ": pipe \"%s\"'s read end already closed\n",pp->pname);
						return 1;
					}
				} else {
					(void) close(pp->d[0]);
					desc_used[pp->d[0]] = 0;
				}
				pp->d[0] = -1;
			}
			if (which & 2) {
				if (pp->d[1] == -1) {
					if ((which & 4) == 0) {
						desc_printf(2,SHNAME ": pipe \"%s\"'s write end already closed\n",pp->pname);
						return 1;
					}
				} else {
					(void) close(pp->d[1]);
					desc_used[pp->d[1]] = 0;
				}
				pp->d[1] = -1;
			}
			if (pp->d[0] == -1 && pp->d[1] == -1) {
				*ppp = pp->next;
				free(pp);
			}
			return 0;
		}
	}
	desc_printf(2,SHNAME ": shell pipe \"%s\" not found\n",dynbuf_get_elt(&cmdp->words,1));
	return 1;
}

struct builtin {
	char	*cmd;
	int	(*fn)(struct cmd	*cmdp); /* returns "exit status" */
} builtins[] = {
	{ "cd",		do_cd, },
	{ "chdir",	do_cd, },
	{ "exit",	do_exit, },
	{ "pipe",	do_pipe, },
	{ "pclose",	do_pclose, },
};

int	do_builtins(struct cmd	*cmdp)
{
	int	i, estatus;
	char	nbuf[16];

	for (i = 0; i < sizeof builtins / sizeof builtins[0]; i++) {
		if (!strcmp(dynbuf_get_elt(&cmdp->words,0),builtins[i].cmd)) {
			estatus = (*builtins[i].fn)(cmdp);
			sprintf(nbuf,"%d",estatus);
			dollar_set_var(dupstr("?"),dupstr(nbuf));
			return 1;
		}
	}
	return 0;
}

void	handle_io_redir(struct redir_op	*rop)
{
	int		d;
	struct plist	*pp;

	if (rop) {
		handle_io_redir(rop->next);
		switch (rop->src_type) {
		case rnone:
			printf("SHOULD NOT HAPPEN\n");
			abort();
			break;
		case rfile:
			switch (rop->dir) {
			case rfappend:
				d = open(rop->src.name,O_APPEND|O_CREAT|O_WRONLY,0777);
				if (d == -1) {
					perror(SHNAME);
					desc_printf(2,"Cannot open append file %s\n",rop->src.name);
					_exit(1);
				}
				if (d != rop->target_desc) {
					desc_check(rop->target_desc);
					(void) dup2(d,rop->target_desc);
					(void) close(d);
				}
				break;
			case rout:
				d = open(rop->src.name,O_TRUNC|O_CREAT|O_WRONLY,0777);
				if (d == -1) {
					perror(SHNAME);
					desc_printf(2,"Cannot open overwrite file %s\n",rop->src.name);
					_exit(1);
				}
				if (d != rop->target_desc) {
					desc_check(rop->target_desc);
					(void) dup2(d,rop->target_desc);
					(void) close(d);
				}
				break;
			case rin:
				d = open(rop->src.name,O_RDONLY,0777);
				if (d == -1) {
					perror(SHNAME);
					desc_printf(2,"Cannot open read file %s\n",rop->src.name);
					_exit(1);
				}
				if (d != rop->target_desc) {
					desc_check(rop->target_desc);
					(void) dup2(d,rop->target_desc);
					(void) close(d);
				}
				break;
			}
			break;
		case rdesc:
			desc_check(rop->target_desc);
			(void) dup2(rop->src.desc,rop->target_desc);
			break;
		case rpipe:
			for (pp = shpipes; pp; pp = pp->next) {
				if (!strcmp(pp->pname,rop->src.name)) {
					if (rop->dir == rin) {
						if (pp->d[0] != -1 && pp->d[0] != rop->target_desc) {
							desc_check(rop->target_desc);
							dup2(pp->d[0],rop->target_desc);
						} else {
							desc_printf(2,SHNAME ": shell pipe \"%s\"'s read end already closed\n");
							_exit(1);
						}
					} else {
						if (pp->d[1] != -1 && pp->d[1] != rop->target_desc) {
							desc_check(rop->target_desc);
							dup2(pp->d[1],rop->target_desc);
						} else {
							desc_printf(2,SHNAME ": shell pipe \"%s\"'s write end already closed\n");
							_exit(1);
						}

					}
					goto dbl_break;
				}
			}
			desc_printf(2,SHNAME ": shell pipe \"%s\" not found\n",rop->src.name);
			_exit(1);
			break;
		case rclose:
			/*
			 * XXX -- users should not be referring to our
			 * pipe descriptors
			 */
			desc_check(rop->target_desc);
			(void) close(rop->target_desc);
			break;
		case rboth:
			/*
			 * there is >& syntax, but no >>& syntax
			 */
			d = open(rop->src.name,O_TRUNC|O_CREAT|O_WRONLY,0777);
			if (d == -1) {
				perror(SHNAME);
				desc_printf(2,"Cannot open overwrite file %s\n",rop->src.name);
				_exit(1);
			}
			/*
			 * XXX -- rop->target_desc is IGNORED
			 * also, we assume 1 and 2 are never pipe descriptors.
			 */
			if (d != 1) (void) dup2(d,1);
			if (d != 2) (void) dup2(d,2);
			if (d != 1 && d != 2) (void) close(d);

			break;
		}
	dbl_break:
		;
	}
}

void	do_varset(char		*setcmd,
		  int		exportp)
{
	char	*eq, *n;

	eq = strchr(setcmd,'=');
	if (!eq) {
		desc_printf(2,SHNAME ": do_varset, internal error: \"%s\" has no equal sign\n",setcmd);
		abort();
	}
	/* hack */
	*eq = '\0';
	dollar_set_var(n = dupstr(setcmd),dupstr(eq+1));
	*eq = '=';
	if (exportp) dollar_mark_exported(n);
}

void	run_command(struct cmd	*cmdp)
{
	int		nelt;
	int		i, nsetcmd;
	int		pid, wstatus;
	char		*cmdname, **vp, nbuf[8];

	nelt = dynbuf_nelt(&cmdp->words);
	nsetcmd = dynbuf_nelt(&cmdp->vars);

	if (!nelt) {
		if (debug) {
			if (nsetcmd) printf("Setting shell variables in parent\n");
			else printf("Null command\n");
		}
		for (i = 0; i < nsetcmd; i++) {
			do_varset((char *) dynbuf_get_elt(&cmdp->vars,i),0);
		}
		goto cleanup;
	}
	if (debug) {
		printf("Command: %s has %d variable setting commands\n",(char *) dynbuf_get_elt(&cmdp->words,0),nsetcmd);
		for (i = 0; i < nsetcmd; i++) {
			printf("%s\n",(char *) dynbuf_get_elt(&cmdp->vars,i));
		}
		printf("Command: %s has %d arguments in its arglist\n",(char *) dynbuf_get_elt(&cmdp->words,0),nelt);
		for (i = 0; i < nelt; i++) {
			printf("argv[%d] = \"%s\"\n",i,(char *) dynbuf_get_elt(&cmdp->words,i));
		}
		printf("I/O redirection:\n");
		show_rops(cmdp->rop);
		printf("Command runs in the %sground\n\n",cmdp->background ? "back" : "fore");
	}
	if (do_builtins(cmdp)) {
		/* "var=value cd .." has no effect in bash either */
		goto cleanup;
	}

	switch (pid = fork()) {
	case -1:
		perror(SHNAME ": run_command: fork");
		desc_printf(2,SHNAME ": cannot fork subprocess to run command %s\n",(char *) dynbuf_get_elt(&cmdp->words,0));
		goto cleanup;
	case 0:
		/* handle all I/O redirections */
		handle_io_redir(cmdp->rop);
		/* export setting commands into environment */
		for (i = 0; i < nsetcmd; i++) {
			do_varset((char *) dynbuf_get_elt(&cmdp->vars,i),1);
		}
		environ = dollar_construct_envp();
		/* close pipe descriptors */
		pipe_cleanup();
		/* construct arglist */
		cmd_add_word(cmdp,0);
		vp = (char **) dynbuf_take_buf(&cmdp->words);
		cmdname = vp[0];
		execvp(cmdname, vp);
		desc_printf(2,"Command %s not found\n",vp[0]);
		_exit(1);
	default:
		if (!cmdp->background) {
			waitpid(pid,&wstatus,0);
			if (WIFSIGNALED(wstatus)) {
				desc_printf(2,"Command %s killed by signal %d\n",(char *) dynbuf_get_elt(&cmdp->words,0),WTERMSIG(wstatus));
			} else {
				sprintf(nbuf,"%d",WEXITSTATUS(wstatus));
				dollar_set_var(dupstr("?"),dupstr(nbuf));
				if (WEXITSTATUS(wstatus) != 0) {
					desc_printf(2,"Command %s exited with status %d\n",(char *) dynbuf_get_elt(&cmdp->words,0),WEXITSTATUS(wstatus));
				}
			}
		}
	}
 cleanup:
	;
}
